/*
 * Copyright 2019 European Commission
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they will be approved by the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 *     https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and limitations under the Licence.
 */
package eu.europa.ec.leos.services.support.xml;

import eu.europa.ec.leos.domain.cmis.Content;
import eu.europa.ec.leos.domain.cmis.common.VersionType;
import eu.europa.ec.leos.domain.cmis.document.Bill;
import eu.europa.ec.leos.domain.cmis.document.XmlDocument;
import eu.europa.ec.leos.domain.cmis.metadata.BillMetadata;
import eu.europa.ec.leos.i18n.LanguageHelper;
import eu.europa.ec.leos.services.store.WorkspaceService;
import eu.europa.ec.leos.services.support.xml.ref.LabelArticleElementsOnly;
import eu.europa.ec.leos.services.support.xml.ref.LabelArticlesOrRecitalsOnly;
import eu.europa.ec.leos.services.support.xml.ref.LabelCitationsOnly;
import eu.europa.ec.leos.services.support.xml.ref.LabelHigherOrderElementsOnly;
import eu.europa.ec.leos.services.support.xml.ref.Ref;
import eu.europa.ec.leos.test.support.LeosTest;
import io.atlassian.fugue.Option;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import java.time.Instant;
import java.util.Arrays;
import java.util.Locale;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class ReferenceLabelServiceTest extends LeosTest {

    protected byte[] docContent;

    protected XmlDocument document;

    @InjectMocks
    protected ReferenceLabelServiceImplForProposal referenceLabelGenerator = new ReferenceLabelServiceImplForProposal();

    @Mock
    protected LanguageHelper languageHelper;

    @Mock
    private WorkspaceService workspaceService;

    @Before
    public void setup() {
        super.setup();

        docContent = doc.getBytes(UTF_8);
        ReflectionTestUtils.setField(referenceLabelGenerator, "labelHandlers",
                Arrays.asList(new LabelHigherOrderElementsOnly(),
                        new LabelArticlesOrRecitalsOnly(),
                        new LabelCitationsOnly(),
                        new LabelArticleElementsOnly()));

        when(languageHelper.getCurrentLocale()).thenReturn(new Locale("en"));

        Content content = mock(Content.class);
        Content.Source source = mock(Content.Source.class);
        when(source.getBytes()).thenReturn(docContent);
        when(content.getSource()).thenReturn(source);
        document = getMockedBill(content);
        when(workspaceService.findDocumentByRef("bill", XmlDocument.class)).thenReturn(document);
    }

    /**
     * The aim of the test is not to compare the result rather than to check if NullPointerException occurs.
     */
    @Test
    public void generateLabel_notSupportedElement_shouldNotThrowNPE() throws Exception {
        referenceLabelGenerator.generateLabel(Arrays.asList(new Ref("","preface", "")), "", "a5_FeJW6z", document.getContent().get().getSource().getBytes());
        referenceLabelGenerator.generateLabel(Arrays.asList(new Ref("","preamble", "")), "", "a5_FeJW6z",  document.getContent().get().getSource().getBytes());
        referenceLabelGenerator.generateLabel(Arrays.asList(new Ref("","cits", "")), "", "a5_FeJW6z", document.getContent().get().getSource().getBytes());
        referenceLabelGenerator.generateLabel(Arrays.asList(new Ref("","recs", "")), "", "a5_FeJW6z", document.getContent().get().getSource().getBytes());
    }

    protected static final String doc = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                    + "<akomaNtoso xmlns=\"http://docs.oasis-open.org/legaldocml/ns/akn/3.0\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">"
                    + "    <bill name=\"regulation\">"
                    + "        <coverPage xml:id=\"coverpage\">"
                    + "            <container name=\"institutionOwnerLogo\" xml:id=\"coverpage__container_1\">"
                    + "                <p>bla</p>"
                    + "            </container>"
                    + "            <longTitle xml:id=\"coverpage__longtitle\">"
                    + "                <p xml:id=\"coverpage__longtitle__p\">"
                    + "                    <docStage xml:id=\"coverpage__longtitle__p__docstage\">Proposal for a</docStage>"
                    + "                    <docType refersTo=\"~REG\" xml:id=\"coverpage__longtitle__p__doctype\">REGULATION OF THE EUROPEAN PARLIAMENT AND OF THE COUNCIL</docType>"
                    + "                    <docPurpose xml:id=\"coverpage__longtitle__p__docpurpose\">on Test</docPurpose>"
                    + "                </p>"
                    + "            </longTitle>"
                    + "            <container name=\"docLabel\" xml:id=\"coverpage__container_2\">"
                    + "                <p xml:id=\"akn_mPCLZl\" ></p>"
                    + "            </container>"
                    + "            <container name=\"language\" xml:id=\"coverpage__container_3\" refersTo=\"#frbrLanguage\">"
                    + "                <p xml:id=\"coverpage__container_3__p\">EN</p>"
                    + "            </container>"
                    + "        </coverPage>"
                    + "        <preface xml:id=\"preface\">"
                    + "            <longTitle xml:id=\"preface__longtitle\">"
                    + "                <p xml:id=\"preface__longtitle__p\">"
                    + "                    <docStage xml:id=\"preface__longtitle__p__docstage\">Proposal for a</docStage>"
                    + "                    <docType refersTo=\"~REG\" xml:id=\"preface__longtitle__p__doctype\">REGULATION OF THE EUROPEAN PARLIAMENT AND OF THE COUNCIL</docType>"
                    + "                    <docPurpose xml:id=\"preface__longtitle__p__docpurpose\">on Test</docPurpose>"
                    + "                </p>"
                    + "            </longTitle>"
                    + "        </preface>"
                    + "        <preamble xml:id=\"preamble\">"
                    + "            <formula xml:id=\"preamble__formula_1\" name=\"actingEntity\">"
                    + "                <p xml:id=\"preamble__formula_1__p\">THE EUROPEAN PARLIAMENT AND THE COUNCIL OF THE EUROPEAN UNION,</p>"
                    + "            </formula>"
                    + "            <citations xml:id=\"cits\" leos:editable=\"false\">"
                    + "                <citation refersTo=\"~legalBasis\" xml:id=\"cit_1\" leos:editable=\"true\">"
                    + "                    <p xml:id=\"cit_1__p\">Having regard to the Treaty on the Functioning of the European Union, and in particular Article [...] thereof,</p>"
                    + "                </citation>"
                    + "                <citation refersTo=\"~legalBasis\" xml:id=\"cit_2\" leos:editable=\"true\">"
                    + "                    <p xml:id=\"cit_2__p\">Having regard to the proposal from the European Commission,</p>"
                    + "                </citation>"
                    + "                <citation refersTo=\"~legalBasis\" xml:id=\"cit_3\" leos:editable=\"true\">"
                    + "                    <p xml:id=\"cit_3__p\">After transmission of the draft legislative act to the national Parliaments,</p>"
                    + "                </citation>"
                    + "                <citation refersTo=\"~legalBasis\" xml:id=\"cit_4\" leos:editable=\"true\">"
                    + "                    <p xml:id=\"cit_4__p\">Having regard to the opinion of the European Economic and Social Committee<authorialNote xml:id=\"authorialnote_1\" marker=\"1\" placement=\"bottom\">"
                    + "                        <p xml:id=\"authorialNote_1__p\">OJ C [...], [...], p. [...]</p>"
                    + "                    </authorialNote>,</p>"
                    + "                </citation>"
                    + "            </citations>"
                    + "            <recitals xml:id=\"recs\" leos:editable=\"false\">"
                    + "                <intro xml:id=\"recitalsIntro\"><p xml:id=\"recs_intro\">Whereas:</p></intro>"
                    + "                <recital xml:id=\"rec_1\" leos:editable=\"true\">"
                    + "                    <num xml:id=\"rec_1__num\">(1)</num>"
                    + "                    <p xml:id=\"rec_1__p\">Recital...</p>"
                    + "                </recital>"
                    + "                <recital xml:id=\"rec_2\" leos:editable=\"true\">"
                    + "                    <num xml:id=\"rec_2__num\">(2)</num>"
                    + "                    <p xml:id=\"rec_2__p\">Recital...</p>"
                    + "                </recital>"
                    + "                <recital xml:id=\"rec_3\" leos:editable=\"true\">"
                    + "                    <num xml:id=\"rec_3__num\">(3)</num>"
                    + "                    <p xml:id=\"rec_3__p\">Recital...</p>"
                    + "                </recital>"
                    + "                <recital xml:id=\"rec_4\" leos:editable=\"true\">"
                    + "                    <num xml:id=\"rec_4__num\">(4)</num>"
                    + "                    <p xml:id=\"rec_4__p\">Recital...</p>"
                    + "                </recital>"
                    + "            </recitals>"
                    + "            <formula xml:id=\"preamble__formula_2\" name=\"adoption\">"
                    + "                <p xml:id=\"preamble__formula_2__p\">HAVE ADOPTED THIS REGULATION:</p>"
                    + "            </formula>"
                    + "        </preamble>"
                    + "        <body xml:id=\"body\">"
                    + "             <part xml:id=\"part11\">"
                    + "                <num xml:id=\"part1n1\" class=\"PartNumber\">Part XI</num>"
                    + "                <heading xml:id=\"part1h1\" class=\"PartHeading\">FINAL PROVISIONS</heading>"
                    + "                <article xml:id=\"a1\">"
                    + "                    <num xml:id=\"a1n1\" class=\"ArticleNumber\">Article 1</num>"
                    + "                    <paragraph xml:id=\"a1p1\">"
                    + "                        <num  xml:id=\"a1p1n1\" class=\"Paragraph(numbered)\">1.</num>"
                    + "                        <content xml:id=\"a1p1c1\" >"
                    + "                            <p  xml:id=\"a1p1c1p1\" class=\"Paragraph(numbered)\">Subject to paragraph 2, this Regulation<authorialNote marker=\"2\" xml:id=\"a1au1\"><p xml:id=\"ptest1\">TestNote1</p></authorialNote> shall apply from 1 January 2013.</p>"
                    + "                        </content>"
                    + "                    </paragraph>"
                    + "                    <paragraph xml:id=\"a1p2\">"
                    + "                        <num  xml:id=\"a1p2n1\" class=\"Paragraph(numbered)\">2.</num>"
                    + "                        <content xml:id=\"a1p2c1\">"
                    + "                            <p  xml:id=\"a1p2c1p1\" class=\"Paragraph(numbered)\">Article 436(1) shall apply from 1 January 2015<authorialNote marker=\"4\" xml:id=\"a2a2\"><p xml:id=\"ptest2\">TestNote2</p></authorialNote>.</p>"
                    + "                        </content>"
                    + "                    </paragraph>"
                    + "                    <paragraph xml:id=\"a1p3\">"
                    + "                        <num  xml:id=\"a1p3n1\" class=\"Paragraph(numbered)\">3.</num>"
                    + "                        <content xml:id=\"a1p3c1\">"
                    + "                            <p  xml:id=\"a1p3c1p1\" class=\"Paragraph(numbered)\">Article 436(1) shall apply from 1 January 2015.</p>"
                    + "                        </content>"
                    + "                    </paragraph>"
                    + "                </article>"
                    + "                <article xml:id=\"a2\">"
                    + "                    <num  xml:id=\"a2n3\" class=\"ArticleNumber\">Article 20</num>"
                    + "                    <alinea xml:id=\"a2aal1\">"
                    + "                        <content xml:id=\"a2c1\">"
                    + "                            <p xml:id=\"a2c1p1\" class=\"Paragraph(unnumbered)\">This Regulation shall enter into force on the day following that of its publication in the <i xml:id=\"i2\">Official Journal of the European<authorialNote marker=\"8\" xml:id=\"a2au3\"><p>TestNote3</p></authorialNote> Union</i>.</p>"
                    + "                        </content>"
                    + "                    </alinea>"
                    + "                </article>"
                    + "            </part>"
                    + "             <part xml:id=\"part22\">"
                    + "                <num xml:id=\"part2n1\" class=\"PartNumber\">Part XII</num>"
                    + "                <heading xml:id=\"part2h1\" class=\"PartHeading\">POST  PROVISIONS</heading>"
                    + "                <article xml:id=\"a3\">"
                    + "                    <num  xml:id=\"a3n3\" class=\"ArticleNumber\">Article 21</num>"
                    + "                    <alinea xml:id=\"a3aal1\">"
                    + "                        <content xml:id=\"a3c1\">"
                    + "                            <p xml:id=\"a3c1p1\" class=\"Paragraph(unnumbered)\">This Regulation shall enter into force on the day following that of its publication in the <i xml:id=\"i2\">Official Journal of the European<authorialNote marker=\"8\" xml:id=\"a3\"><p>TestNote3</p></authorialNote> Union</i>.</p>"
                    + "                        </content>"
                    + "                    </alinea>"
                    + "                </article>"
                    + "             </part>"
                    + "             <part xml:id=\"part3\">"
                    + "                <num xml:id=\"part3n1\" class=\"PartNumber\">Part XIV</num>"
                    + "                <heading xml:id=\"part3h1\" class=\"PartHeading\">POST FINAL PROVISIONS</heading>"
                    + "                 <section xml:id=\"p3s1\">"
                    + "                    <num xml:id=\"part3n1\" class=\"PartNumber\">Section X</num>"
                    + "                    <heading xml:id=\"part3h1\" class=\"PartHeading\">POST FINAL PROVISIONS</heading>"
                    + "                    <article xml:id=\"a4\">"
                    + "                        <num  xml:id=\"a4n3\" class=\"ArticleNumber\">Article 45</num>"
                    + "                        <paragraph xml:id=\"art_1_i0IkfN\">"
                    + "                             <num xml:id=\"art_1_uB6sEl\">1.</num>"
                    + "                            <subparagraph xml:id=\"art_1_D48DVp\">"
                    + "                                <content xml:id=\"art_1_bFb2ix\">"
                    + "                                    <p xml:id=\"art_1_iXITCG\">Lorem ipsum dolor sit amet, consectetur adipiscing elit. Mauris libero odio, suscipit eu tincidunt at, viverra sit amet nisl."
                    + "                                    </p>"
                    + "                                 </content>"
                    + "                             </subparagraph>"
                    + "                    <list xml:id=\"art_1_bUwn2R\">"
                    + "                        <point xml:id=\"art_1_4T9QRZ\">"
                    + "                            <num xml:id=\"art_1_VNANTx\">(a)</num>"
                    + "                            <alinea xml:id=\"art_1_pQXVCt\">"
                    + "                                <content xml:id=\"art_1_CUCaDf\">"
                    + "                                    <p xml:id=\"art_1_aW5OHt\">Quisque quis est tortor. Etiam ac lectus in eros mollis mollis nec et lacus. Duis non arcu ut mi volutpat feugiat."
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </alinea>"
                    + "                            <list xml:id=\"art_1_4DUkpJ\">"
                    + "                                <point xml:id=\"art_1_w74QP2\">"
                    + "                                    <num xml:id=\"art_1_GMlAFk\">(i)</num>"
                    + "                                    <content xml:id=\"art_1_l3LI05\">"
                    + "                                        <p xml:id=\"art_1_zpb87z\">Nam venenatis sapien tellus, sed iaculis est facilisis non. Sed ut elit erat. Interdum et malesuada fames ac ante ipsum primis in faucibus."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                                <point xml:id=\"art_1_PcA2LM\">"
                    + "                                    <num xml:id=\"art_1_2ewtEH\">(ii)</num>"
                    + "                                    <content xml:id=\"art_1_KhZCZd\">"
                    + "                                        <p xml:id=\"art_1_GFgFMf\">Suspendisse ac ante sit amet magna consectetur semper vehicula in tellus. Suspendisse non mollis dolor."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                            </list>"
                    + "                        </point>"
                    + "                        <point xml:id=\"art_1_1uV2VS\">"
                    + "                            <num xml:id=\"art_1_N5soNE\">(b)</num>"
                    + "                            <content xml:id=\"art_1_teYKVT\">"
                    + "                                <p xml:id=\"art_1_1Det6t\">Donec quis est sapien. Donec efficitur nulla non erat egestas consectetur eu eu sem."
                    + "                                </p>"
                    + "                            </content>"
                    + "                        </point>"
                    + "                    </list>"
                    + "                </paragraph>"
                    + "                <paragraph xml:id=\"art_1_TxunI0\">"
                    + "                    <num xml:id=\"art_1_ad3Rbl\">2.</num>"
                    + "                    <content xml:id=\"art_1_fISEsZ\">"
                    + "                        <p xml:id=\"art_1_oxdTif\">Pellentesque congue orci ac tincidunt ultrices. Mauris volutpat quam ut feugiat auctor. Phasellus eros est, pharetra nec sollicitudin commodo, dictum quis ligula."
                    + "                        </p>"
                    + "                    </content>"
                    + "                </paragraph>"
                    + "                <paragraph xml:id=\"art_1_jqI9E1\">"
                    + "                    <num xml:id=\"art_1_ORVJhj\">3.</num>"
                    + "                    <subparagraph xml:id=\"art_1_8BEsrS\">"
                    + "                        <content xml:id=\"art_1_UyWhui\">"
                    + "                            <p xml:id=\"art_1_TKV1Yb\">Nullam commodo auctor arcu, quis fringilla metus malesuada convallis. Pellentesque placerat justo in malesuada scelerisque. Mauris vel lorem nulla. Mauris nunc est, tincidunt ac justo sed, molestie aliquam lacus."
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </subparagraph>"
                    + "                    <list xml:id=\"art_1_ViIaZT\">"
                    + "                        <point xml:id=\"art_1_jMGiAd\">"
                    + "                            <num xml:id=\"art_1_0PiSVo\">(a)</num>"
                    + "                            <alinea xml:id=\"art_1_w3g0rn\">"
                    + "                                <content xml:id=\"art_1_LzQCtQ\">"
                    + "                                    <p xml:id=\"art_1_HnxDcU\">Nulla facilisi. In porta ultricies orci non sollicitudin. Curabitur gravida neque eu sem vestibulum, sed gravida ex ultricies."
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </alinea>"
                    + "                            <list xml:id=\"art_1_hvcWBw\">"
                    + "                                <point xml:id=\"art_1_OrhWbv\">"
                    + "                                    <num xml:id=\"art_1_2DSp05\">(i)</num>"
                    + "                                    <content xml:id=\"art_1_VkisV8\">"
                    + "                                        <p xml:id=\"art_1_MzfGNC\">Duis erat neque, consectetur eu odio nec, suscipit euismod orci. Mauris euismod interdum malesuada."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                                <point xml:id=\"art_1_Uxo4c1\">"
                    + "                                    <num xml:id=\"art_1_GevAET\">(ii)</num>"
                    + "                                    <alinea xml:id=\"art_1_jM5Tig\">"
                    + "                                        <content xml:id=\"art_1_ezShzW\">"
                    + "                                            <p xml:id=\"art_1_Rjqlk5\">asdadasd</p>"
                    + "                                        </content>"
                    + "                                    </alinea>"
                    + "                                    <list xml:id=\"art_1_iWQoV4\">"
                    + "                                        <point xml:id=\"art_1_l3MS0E\">"
                    + "                                            <num xml:id=\"art_1_IJgtM3\">-</num>"
                    + "                                            <alinea xml:id=\"art_1_5wNT45\">"
                    + "                                                <content xml:id=\"art_1_FCHhwd\">"
                    + "                                                    <p xml:id=\"art_1_HGUYv5\">Ut eget massa quis erat pulvinar finibus sed pretium sem. Sed id nunc ac odio posuere hendrerit sed ac tellus. Aenean dapibus dignissim diam, in laoreet risus condimentum a. Aliquam tincidunt tellus non enim maximus blandit."
                    + "                                                    </p>"
                    + "                                                </content>"
                    + "                                            </alinea>"
                    + "                                            <list xml:id=\"art_1_vU7GDV\">"
                    + "                                                <point xml:id=\"art_1_DBRTQn\">"
                    + "                                                    <num xml:id=\"art_1_YUGkcC\">-</num>"
                    + "                                                    <content xml:id=\"art_1_MgBQ2G\">"
                    + "                                                        <p xml:id=\"art_1_KG1KdL\">"
                    + "                                                            Pellentesque accumsan dapibus velit, ac malesuada sapien aliquet ut. Aliquam quis enim non dui sodales aliquet. In sed ante in ligula blandit malesuada. Maecenas nec metus iaculis, porta ex id, mollis felis."
                    + "                                                        </p>"
                    + "                                                    </content>"
                    + "                                                </point>"
                    + "                                            </list>"
                    + "                                        </point>"
                    + "                                    </list>"
                    + "                                </point>"
                    + "                                <point xml:id=\"art_1_CY6Nsa\">"
                    + "                                    <num xml:id=\"art_1_ETB4nx\">(iii)</num>"
                    + "                                    <content xml:id=\"art_1_O2aavx\">"
                    + "                                        <p xml:id=\"art_1_7o6VRy\">Sed sem magna, cursus et commodo ultricies, ultrices sed libero. Ut vel ex odio. Suspendisse viverra tempor sapien, facilisis suscipit quam porttitor in."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                            </list>"
                    + "                        </point>"
                    + "                        <point xml:id=\"art_1_Orvvv\">"
                    + "                              <num xml:id=\"art_1_2Dcccc\">(b)</num>"
                    + "                              <content xml:id=\"art_1_Vkcc8\">"
                    + "                                   <p xml:id=\"art_1_Mccc\">Duis erat neque, consectetur eu odio nec, suscipit euismod orci. Mauris euismod interdum malesuada."
                    + "                                   </p>"
                    + "                              </content>"
                    + "                        </point>"
                    + "                    </list>"
                    + "                </paragraph>"
                    + "                <paragraph xml:id=\"art_1_kpm2jb\">"
                    + "                    <num xml:id=\"art_1_H1XU47\">4.</num>"
                    + "                    <subparagraph xml:id=\"art_1_S9i3Fz\">"
                    + "                        <content xml:id=\"art_1_oczbxe\">"
                    + "                            <p xml:id=\"art_1_eyHMBk\">Nulla ut odio rutrum, tempus lorem a, tincidunt nisl. Fusce convallis pellentesque nisi, sit amet iaculis neque pharetra ut. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere cubilia Curae;"
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </subparagraph>"
                    + "                    <list xml:id=\"art_1_kloBJf\">"
                    + "                        <point xml:id=\"art_1_vZe95X\">"
                    + "                            <num xml:id=\"art_1_2TXvp3\">(a)</num>"
                    + "                            <content xml:id=\"art_1_2UKqwj\">"
                    + "                                <p xml:id=\"art_1_elCZdB\">Phasellus efficitur viverra ex, maximus volutpat erat feugiat ac. Cras augue nunc, elementum vitae porttitor sit amet, posuere nec ex."
                    + "                                </p>"
                    + "                            </content>"
                    + "                        </point>"
                    + "                        <point xml:id=\"art_1_830kLA\">"
                    + "                            <num xml:id=\"art_1_gkz6kf\">(b)</num>"
                    + "                            <alinea xml:id=\"art_1_Cfr97w\">"
                    + "                                <content xml:id=\"art_1_xmNDzv\">"
                    + "                                    <p xml:id=\"art_1_JDj92t\">Vestibulum egestas, ante dignissim auctor vulputate, turpis nisi pulvinar enim, molestie lacinia dolor justo nec leo."
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </alinea>"
                    + "                            <list xml:id=\"art_1_rUHNOL\">"
                    + "                                <point xml:id=\"art_1_hZrAiD\">"
                    + "                                    <num xml:id=\"art_1_cQhlH3\">(i)</num>"
                    + "                                    <content xml:id=\"art_1_UsubFK\">"
                    + "                                        <p xml:id=\"art_1_rSidlP\">Phasellus ullamcorper, tellus ac pellentesque volutpat, neque eros placerat erat, eu suscipit tellus lacus at risus. I"
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                                <point xml:id=\"art_1_NibNsa\">"
                    + "                                    <num xml:id=\"art_1_2daSks\">(ii)</num>"
                    + "                                    <content xml:id=\"art_1_8Z1O9o\">"
                    + "                                        <p xml:id=\"art_1_GTysXx\">nteger nec diam enim. Duis sit amet maximus elit. Morbi lacus risus, ornare sed tincidunt eget, elementum eu arcu."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </point>"
                    + "                            </list>"
                    + "                        </point>"
                    + "                        <point xml:id=\"art_1_CkbJ1I\">"
                    + "                            <num xml:id=\"art_1_9jmaxo\">(c)</num>"
                    + "                            <content xml:id=\"art_1_vqv65v\">"
                    + "                                <p xml:id=\"art_1_CB3nb2\">Vestibulum ac neque tempor, pellentesque arcu accumsan, blandit metus. Proin luctus justo vitae lectus congue tempus."
                    + "                                </p>"
                    + "                            </content>"
                    + "                        </point>"
                    + "                    </list>"
                    + "                </paragraph>"
                    + "                <paragraph xml:id=\"art_1_A42pW6\">"
                    + "                    <num xml:id=\"art_1_AYrfwX\">5.</num>"
                    + "                    <content xml:id=\"art_1_vvi8Re\">"
                    + "                        <p xml:id=\"art_1_m9ZMyE\">Donec consequat neque aliquet, scelerisque dolor sit amet, vehicula enim. Morbi ullamcorper ligula ac cursus sagittis. Aenean fringilla enim quis varius molestie. Sed ut massa tortor. Nam bibendum nisl eu mi faucibus, quis aliquet augue molestie. Curabitur venenatis ligula leo, non bibendum est suscipit at."
                    + "                        </p>"
                    + "                    </content>"
                    + "                </paragraph>"
                    + "                <paragraph xml:id=\"art_1_BGt5eN\">"
                    + "                    <num xml:id=\"art_1_5fZk0t\">6.</num>"
                    + "                    <content xml:id=\"art_1_GlIzqV\">"
                    + "                        <p xml:id=\"art_1_HmBP4y\">Etiam placerat, nisl sit amet consequat facilisis, augue mauris tristique orci, vel fermentum odio eros non ipsum. Nunc auctor auctor nibh eu ultrices. Cras nunc odio, varius hendrerit vehicula nec, vehicula a ex."
                    + "                        </p>"
                    + "                    </content>"
                    + "                </paragraph>"
                    + "                     </article>"
                    + "<article xml:id=\"a5\">"
                    + "    <num leos:editable=\"false\" xml:id=\"a4n3\">Article 46</num>"
                    + "    <paragraph xml:id=\"a5_sdplN0\">"
                    + "        <subparagraph xml:id=\"a5_Bxi62k\">"
                    + "            <content xml:id=\"a5_jkVQ6D\">"
                    + "                <p xml:id=\"a5_hR0JEb\">Sed ut perspiciatis unde omnis iste natus error sit voluptatem accusantium doloremque laudantiu."
                    + "                </p>"
                    + "            </content>"
                    + "        </subparagraph>"
                    + "        <list xml:id=\"art_5_bUwn2R\">"
                    + "            <point xml:id=\"a5_fQYvU7\">"
                    + "                <num xml:id=\"a5_N6tAsa\">(a)</num>"
                    + "                <content xml:id=\"a5_SeHdoO\">"
                    + "                    <p xml:id=\"a5_d9rOHd\">Nemo enim ipsam voluptatem quia voluptas sit aspernatur aut odit aut fugit, sed quia consequuntur magni dolores eos qui ratione voluptat"
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a5_lwMbkL\">"
                    + "                <num xml:id=\"a5_3tB2Gi\">(b)</num>"
                    + "                <content xml:id=\"a5_CCcd6s\">"
                    + "                    <p xml:id=\"a5_jl1NNz\">Ut enim ad minima veniam, quis nostrum exercitationem ullam corporis suscipit laboriosam, nisi ut aliquid ex ea commodi consequatur?."
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a5_v0SBeN\">"
                    + "                <num xml:id=\"a5_tBvCXA\">(c)</num>"
                    + "                <content xml:id=\"a5_BRCYbs\">"
                    + "                    <p xml:id=\"a5_O1J7za\">At vero eos et accusamus et iusto odio dignissimos ducimus qui blanditiis praesentium voluptatum deleniti atque corrupti quos dolores et "
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a5_bSsbTj\">"
                    + "                <num xml:id=\"a5_HTftPi\">(d)</num>"
                    + "                <alinea xml:id=\"a5_T0L37f\">"
                    + "                    <content xml:id=\"a5_OxbXVA\">"
                    + "                        <p xml:id=\"a5_as2p0J\">But I must explain to you how all this mistaken idea of denouncing pleasure and praising pain was born and I will "
                    + "                        </p>"
                    + "                    </content>"
                    + "                </alinea>"
                    + "                <list xml:id=\"a5_yOotfQ\">"
                    + "                    <point xml:id=\"a5_TcYL6X\">"
                    + "                        <num xml:id=\"a5_7g9D8K\">(1)</num>"
                    + "                        <content xml:id=\"a5_2bP84D\">"
                    + "                            <p xml:id=\"a5_sBGxr4\">Nor again is there anyone who loves or pursues or desires to obtain pain of itself, because it is pain, but because occasionally"
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </point>"
                    + "	                   <point xml:id=\"a5_Vdotsh\">"
                    + "		                   <num xml:id=\"a5_CI2pyJ\">(2)</num>"
                    + "                        <content xml:id=\"a5_axhVGa\">"
                    + "                            <p xml:id=\"a5_Ezy8hH\">these matters to this principle of selection: he rejects pleasures to secure other g"
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </point>"
                    + "                    <point xml:id=\"a5_jlazVX\">"
                    + "                        <num xml:id=\"a5_WM0bIs\">(3)</num>"
                    + "                        <alinea xml:id=\"a5_nmQFmM\">"
                    + "                            <content xml:id=\"a5_IUNTit\">"
                    + "                                <p xml:id=\"a5_GKGJ61\">Temporibus autem quibusdam et aut officiis debitis aut rerum necessitatibus saepe eveniet ut et voluptates "
                    + "                                </p>"
                    + "                            </content>"
                    + "                        </alinea>"
                    + "                        <list xml:id=\"a5_GEeMl6\">"
                    + "                            <point xml:id=\"a5_HTmAdx\">"
                    + "                                <num xml:id=\"a5_J4IJ2B\">(i)</num>"
                    + "                                <content xml:id=\"a5_330W9E\">"
                    + "                                    <p xml:id=\"a5_II32Py\">Itaque earum rerum hic tenetur a sapiente delectus, ut aut reiciendis voluptatibus maiores alias consequatur aut perferendis doloribus"
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a5_oFgu8x\">"
                    + "                                <num xml:id=\"a5_igzs6w\">(ii)</num>"
                    + "                                <content xml:id=\"a5_BJsv7g\">"
                    + "                                    <p xml:id=\"a5_x84ooC\">On the other hand, we denounce with righteous indignation and dislike men"
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a5_W9FxgJ\">"
                    + "                                <num xml:id=\"a5_rsJfmt\">(iii)</num>"
                    + "                                <content xml:id=\"a5_Qngv1P\">"
                    + "                                    <p xml:id=\"a5_hmGHER\">Et harum quidem rerum facilis est et expedita distinctio. Nam libero tempore, cum soluta nobis est eligendi optio cumque nihil impedit"
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a5_Pxc9VZ\">"
                    + "                                <num xml:id=\"a5_9SRuve\">(iv)</num>"
                    + "                                <alinea xml:id=\"a5_csTSSU\">"
                    + "                                    <content xml:id=\"a5_ScPl1y\">"
                    + "                                        <p xml:id=\"a5_S59ulL\">pain and trouble that are bound to ensue; and equal blame belongs to those who fail "
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </alinea>"
                    + "                                <list xml:id=\"a5_j77BFE\">"
                    + "                                    <point xml:id=\"a5_A8TAMj\">"
                    + "                                        <num xml:id=\"a5_TCRhDa\">-</num>"
                    + "                                        <content xml:id=\"a5_jX0SCD\">"
                    + "                                            <p xml:id=\"a5_r428HZ\">who are so beguiled and demoralized by the charms of pleasure of the moment, so blinded by desire, that they cannot foresee"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                    <point xml:id=\"a5_51QZD5\">"
                    + "                                        <num xml:id=\"a5_0Z3ww0\">-</num>"
                    + "                                        <content xml:id=\"a5_ygopYs\">"
                    + "                                            <p xml:id=\"a5_2bR0wA\">Et harum quidem rerum facilis est et expedita distinctio. Nam libero tempore, cum soluta nobis est eligendi optio"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                    <point xml:id=\"a5_IktngU\">"
                    + "                                        <num xml:id=\"a5_X6Doah\">-</num>"
                    + "                                        <content xml:id=\"a5_c8xcJ6\">"
                    + "                                            <p xml:id=\"a5_XNQCsb\">when our power of choice is untrammelled and when nothing prevents our being able to do what we like best, every pleasure"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                </list>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a5_2Y0ygz\">"
                    + "	                               <num xml:id=\"a5_LIEZmV\">(v)</num>"
                    + "	                               <alinea xml:id=\"a5_JXc2nY\">"
                    + "	                                   <content xml:id=\"a5_Z529Xb\">"
                    + "	                                       <p xml:id=\"a5_IMq3wd\">dummy text of the printing and typesetting industry. Lorem Ipsum has been the industry"
                    + "	                                       </p>"
                    + "	                                   </content>"
                    + "	                               </alinea>"
                    + "	                               <list xml:id=\"a5_SdkYgf\">"
                    + "		                               <point xml:id=\"a5_nirQOI\">"
                    + "			                               <num xml:id=\"a5_wfNLqT\">-</num>"
                    + "                                        <content xml:id=\"a5_UBYG1h\">"
                    + "				                               <p xml:id=\"a5_N2zL23\">when an unknown printer took a galley of type and scrambled it to make a type specimen book"
                    + "				                               </p>"
                    + "			                               </content>"
                    + "		                               </point>"
                    + "	                                   <point xml:id=\"a5_4Lmhzh\">"
                    + "	                                       <num xml:id=\"a5_ViBOlt\">-</num>"
                    + "	                                       <content xml:id=\"a5_NflPYn\">"
                    + "	                                           <p xml:id=\"a5_KRBtGL\"> It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum "
                    + "	                                           </p>"
                    + "	                                       </content>"
                    + "	                                   </point>"
                    + "	                               </list>"
                    + "	                           </point>"
                    + "                        </list>"
                    + "                    </point>"
                    + "                </list>"
                    + "            </point>"
                    + "        </list>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"a5_FeJW6z\">"
                    + "        <content xml:id=\"a5_wLK2xv\">"
                    + "            <p xml:id=\"a5_eDPBGR\">Quis autem</p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"art_5_TxunI0\">"
                    + "        <content xml:id=\"art_5_fISEsZ\">"
                    + "            <p xml:id=\"art_5_oxdTif\">Pellentesque congue orci ac tincidunt ultrices. Mauris volutpat quam ut feugiat</p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"art_5_A42pW6\">"
                    + "        <content xml:id=\"art_5_vvi8Re\">"
                    + "            <p xml:id=\"art_5_m9ZMyE\">Donec consequat neque aliquet, scelerisque dolor sit amet, vehicula enim. Mor</p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"art_5_BGt5eN\">"
                    + "        <content xml:id=\"art_5_GlIzqV\">"
                    + "            <p xml:id=\"art_5_HmBP4y\">Etiam placerat, nisl sit amet consequat facilisis, augue mauris tristique orci"
                    + "            </p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "</article>"


                    + "<article xml:id=\"a6\" leos:editable=\"true\" leos:deletable=\"true\">"
                    + "    <num leos:editable=\"false\" xml:id=\"a6_XMrJki\">Article 47</num>"
                    + "    <heading xml:id=\"a6_jiDgk5\">Article heading...</heading>"
                    + "    <paragraph xml:id=\"a6_PcpmBy\">"
                    + "        <num xml:id=\"a6_b12NUC\">1.</num>"
                    + "        <subparagraph xml:id=\"a6_9CfL6Y\">"
                    + "            <content xml:id=\"a6_tFHiwl\">"
                    + "                <p xml:id=\"a6_ejVfMJ\">Sed ut perspiciatis unde omnis iste natus error sit voluptatem accusantium doloremque "
                    + "                </p>"
                    + "            </content>"
                    + "        </subparagraph>"
                    + "        <list xml:id=\"a6_jPGdEf\">"
                    + "            <point xml:id=\"a6_uYgixH\">"
                    + "                <num xml:id=\"a6_5yAMw5\">(a)</num>"
                    + "                <content xml:id=\"a6_GuHtk8\">"
                    + "                    <p xml:id=\"a6_5rUuKO\">illo inventore veritatis et quasi architecto beatae vitae dicta sunt explicabo. Nemo enim ipsam voluptatem quia voluptas sit aspernatur aut odit aut fugit"
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a6_fVRJ9F\">"
                    + "                <num xml:id=\"a6_2I0ToA\">(b)</num>"
                    + "                <content xml:id=\"a6_bZ0JfW\">"
                    + "                    <p xml:id=\"a6_j0m0Qh\">ed quia consequuntur magni dolores eos qui ratione voluptatem sequi nesciunt."
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a6_yPOvBG\">"
                    + "                <num xml:id=\"a6_HPkB2E\">(c)</num>"
                    + "                <content xml:id=\"a6_GngLSd\">"
                    + "                    <p xml:id=\"a6_VNqZR3\">Neque porro quisquam est, qui dolorem ipsum quia dolor sit amet, consectetur"
                    + "                    </p>"
                    + "                </content>"
                    + "            </point>"
                    + "            <point xml:id=\"a6_g6qoqW\">"
                    + "                <num xml:id=\"a6_HnR6EF\">(d)</num>"
                    + "                <alinea xml:id=\"a6_MM9gYx\">"
                    + "                    <content xml:id=\"a6_GFAUqN\">"
                    + "                        <p xml:id=\"a6_HqAQFR\">adipisci velit, sed quia non numquam eius modi tempora incidunt ut labore et dolore magnam aliquam quaerat voluptatem"
                    + "                        </p>"
                    + "                    </content>"
                    + "                </alinea>"
                    + "                <list xml:id=\"a6_phLVTu\">"
                    + "                    <point xml:id=\"a6_qJGPtu\">"
                    + "                        <num xml:id=\"a6_qLc8xZ\">(1)</num>"
                    + "                        <content xml:id=\"a6_IE11U5\">"
                    + "                            <p xml:id=\"a6_JStvPl\">Ut enim ad minima veniam, quis nostrum exercitationem ullam corporis suscipit laboriosam, nisi ut aliquid ex ea commodi consequatur? "
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </point>"
                    + "                    <point xml:id=\"a6_JYhj1s\">"
                    + "                        <num xml:id=\"a6_w17jbN\">(2)</num>"
                    + "                        <content xml:id=\"a6_Odh7tU\">"
                    + "                            <p xml:id=\"a6_dvyqzD\">Quis autem vel eum iure reprehenderit qui in ea voluptate velit esse quam nihil molestiae consequatur, vel illum qui dolorem eum fugiat quo voluptas nulla pariatur?"
                    + "                            </p>"
                    + "                        </content>"
                    + "                    </point>"
                    + "                    <point xml:id=\"a6_HVr06Y\">"
                    + "                        <num xml:id=\"a6_zSp8zC\">(3)</num>"
                    + "                        <alinea xml:id=\"a6_6CT2lS\">"
                    + "                            <content xml:id=\"a6_eubdn2\">"
                    + "                                <p xml:id=\"a6_wKYbyY\">But I must explain to you how all this mistaken idea of denouncing pleasure and praising pain was born and I will give you a complete"
                    + "                                </p>"
                    + "                            </content>"
                    + "                        </alinea>"
                    + "                        <list xml:id=\"a6_fGbc93\">"
                    + "                            <point xml:id=\"a6_sdGkGc\">"
                    + "                                <num xml:id=\"a6_Pabtan\">(i)</num>"
                    + "                                <content xml:id=\"a6_WpLOub\">"
                    + "                                    <p xml:id=\"a6_dGJ76Z\">and expound the actual teachings of the great explorer of the truth, the master-builder of human happiness."
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a6_3eqk4L\">"
                    + "                                <num xml:id=\"a6_DraeNG\">(ii)</num>"
                    + "                                <content xml:id=\"a6_GQiCyZ\">"
                    + "                                    <p xml:id=\"a6_JZw1TT\">avoids pleasure itself, because it is pleasure, but because those who do not know how to pursue pleasure rationally encounter "
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a6_dhDm9U\">"
                    + "                                <num xml:id=\"a6_sYPEq3\">(iii)</num>"
                    + "                                <content xml:id=\"a6_0a7rT4\">"
                    + "                                    <p xml:id=\"a6_oWXCg2\">Nor again is there anyone who loves or pursues or desires to obtain pain of itself, because it is pain"
                    + "                                    </p>"
                    + "                                </content>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a6_d42YEi\">"
                    + "                                <num xml:id=\"a6_E4MlYj\">(iv)</num>"
                    + "                                <alinea xml:id=\"a6_grl0Ed\">"
                    + "                                    <content xml:id=\"a6_nfgvVC\">"
                    + "                                        <p xml:id=\"a6_6LW9dQ\">but because occasionally circumstances occur in which toil and pain can procure him some great pleasure."
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </alinea>"
                    + "                                <list xml:id=\"a6_MIvppL\">"
                    + "                                    <point xml:id=\"a6_vFs1j9\">"
                    + "                                        <num xml:id=\"a6_ySGdz9\">-</num>"
                    + "                                        <content xml:id=\"a6_Cyu7dK\">"
                    + "                                            <p xml:id=\"a6_3qb01S\">To take a trivial example, which of us ever undertakes laborious physical exercise"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                    <point xml:id=\"a6_ed2l3M\">"
                    + "                                        <num xml:id=\"a6_qo10DA\">-</num>"
                    + "                                        <content xml:id=\"a6_zLnrlg\">"
                    + "                                            <p xml:id=\"a6_53ZYIP\">except to obtain some advantage from it? But who has any right to find fault with a man who chooses to enjoy a pleasure"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                    <point xml:id=\"a6_H9BTwE\">"
                    + "                                        <num xml:id=\"a6_OepH2p\">-</num>"
                    + "                                        <content xml:id=\"a6_bGZqLT\">"
                    + "                                            <p xml:id=\"a6_ofI0Ga\"> that has no annoying consequences, or one who avoids a pain that produces no resultant pleasure?"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                </list>"
                    + "                            </point>"
                    + "                            <point xml:id=\"a6_GxXx51\">"
                    + "                                <num xml:id=\"a6_TFasEw\">(v)</num>"
                    + "                                <alinea xml:id=\"a6_I5bTCT\">"
                    + "                                    <content xml:id=\"a6_2SermG\">"
                    + "                                        <p xml:id=\"a6_RQUFYN\">vero eos et accusamus et iusto odio dignissimos ducimus qui blanditiis praesentium voluptatum deleniti atque"
                    + "                                        </p>"
                    + "                                    </content>"
                    + "                                </alinea>"
                    + "                                <list xml:id=\"a6_ceBaYq\">"
                    + "                                    <point xml:id=\"a6_GVVgiJ\">"
                    + "                                        <num xml:id=\"a6_Cnxj5F\">-</num>"
                    + "                                        <content xml:id=\"a6_s4Tien\">"
                    + "                                            <p xml:id=\"a6_R51WBA\">corrupti quos dolores et quas molestias"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                    <point xml:id=\"a6_9WhjXw\">"
                    + "                                        <num xml:id=\"a6_cpl34i\">-</num>"
                    + "                                        <content xml:id=\"a6_ok9w1S\">"
                    + "                                            <p xml:id=\"a6_Y2LQsD\">excepturi sint occaecati cupiditate non provident"
                    + "                                            </p>"
                    + "                                        </content>"
                    + "                                    </point>"
                    + "                                </list>"
                    + "                            </point>"
                    + "                        </list>"
                    + "                    </point>"
                    + "                </list>"
                    + "            </point>"
                    + "        </list>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"a6_hJc9Yk\">"
                    + "        <num xml:id=\"a6_dLmLiA\">2.</num>"
                    + "        <content xml:id=\"a6_cRyXmX\">"
                    + "            <p xml:id=\"a6_pWj1hw\">On the other hand, we denounce with righteous indignation and dislike men who are so beguiled and demoralized by the charms of pleasure of the moment"
                    + "            </p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"a6_v4ETw1\">"
                    + "        <num xml:id=\"a6_LMjDMb\">3.</num>"
                    + "        <content xml:id=\"a6_42k6fj\">"
                    + "            <p xml:id=\"a6_GYntlo\">These cases are perfectly simple and easy to distinguish. In a free hour"
                    + "            </p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"a6_h0qOvb\">"
                    + "        <num xml:id=\"a6_obsrG3\">4.</num>"
                    + "        <content xml:id=\"a6_sRd4L5\">"
                    + "            <p xml:id=\"a6_VLWTCp\">When our power of choice is untrammelled and when nothing prevents our being able to do what we like best, every pleasure is to be welcomed and every pain avoided"
                    + "            </p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "    <paragraph xml:id=\"a6_Dd29LR\">"
                    + "        <num xml:id=\"a6_lvCaTN\">5.</num>"
                    + "        <content xml:id=\"a6_5nxuKJ\">"
                    + "            <p xml:id=\"a6_Y9UCiH\">But in certain circumstances and owing to the claims of duty or the obligations of business it will frequently occur that pleasures have to be repudiated"
                    + "            </p>"
                    + "        </content>"
                    + "    </paragraph>"
                    + "</article>"
                    + "                 </section>"
                    + "                  <section xml:id=\"p3s2\">"
                    + "                    <num xml:id=\"part3ns2\" class=\"PartNumber\">Section XI</num>"
                    + "                    <heading xml:id=\"part3h123\" class=\"PartHeading\">POST FINAL PROVISIONS</heading>"
                    + "                 </section>"
                    + "             </part>"
                    + "          </body>"
                    + "        </bill>"
                    + "</akomaNtoso> ";

    private Bill getMockedBill(Content content) {
        BillMetadata billMetadata = new BillMetadata("", "REGULATION", "", "SJ-023", "EN","", "bill", "", "0.1.0");
        return new Bill("1", "Legaltext", "login", Instant.now(), "login", Instant.now(),
                "", "", "Version 1.0.0", "", VersionType.MAJOR, true, "title", null, Arrays.asList(""), Option.some(content), Option.some(billMetadata));
    }
}