/*
 * Copyright 2019 European Commission
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they will be approved by the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 *     https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and limitations under the Licence.
 */
package eu.europa.ec.leos.services.support.xml;

import com.ximpleware.VTDNav;
import com.ximpleware.XMLModifier;
import eu.europa.ec.leos.test.support.LeosTest;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;

import javax.xml.datatype.DatatypeConfigurationException;

import java.io.ByteArrayOutputStream;

import static eu.europa.ec.leos.services.support.xml.XmlHelper.PARAGRAPH;
import static eu.europa.ec.leos.services.support.xml.XmlHelper.POINT;
import static eu.europa.ec.leos.services.support.xml.XmlHelper.SUBPARAGRAPH;
import static eu.europa.ec.leos.services.support.xml.XmlHelper.SUBPOINT;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.mockito.Mockito.when;

public class VtdXmlContentProcessorForMandateTest extends LeosTest {

    @Mock
    SecurityContext securityContext;

    @Mock
    Authentication authentication;

    @Mock
    UserDetails userDetails;

    @InjectMocks
    private VtdXmlContentProcessorForMandate vtdXmlContentProcessor = Mockito
            .spy(new VtdXmlContentProcessorForMandate());

    @Before
    public void onSetUp() throws DatatypeConfigurationException {
        when(userDetails.getUsername()).thenReturn("demo");
        when(authentication.getPrincipal()).thenReturn(userDetails);
        when(securityContext.getAuthentication()).thenReturn(authentication);
        SecurityContextHolder.setContext(securityContext);

        Mockito.doReturn("1900-01-01T00:00:00.000+01:00").when(vtdXmlContentProcessor).getXMLFormatDate();
    }

    @Test
    public void test_updateNewElements_cke_article_ec_paragraph_ec_splitted() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e34277_V6MXDb\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e34277_V6MXDb_MG65v8\" >Article 4</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e34277_V6MXDb_Opeshn\" >Calculating the effect of the use of credit risk mitigation techniques</heading><paragraph xml:id=\"imp_art_d1e34277_V6MXDb_mdQK5X\" leos:origin=\"ec\"><num xml:id=\"imp_art_d1e34277_V6MXDb_rSp9HS\" leos:origin=\"ec\">1.</num><subparagraph xml:id=\"imp_art_d1e34277_V6MXDb_BtcrKZ\" ><content xml:id=\"imp_art_d1e34277_V6MXDb_g2YFMP\" ><p xml:id=\"imp_art_d1e34277_V6MXDb_A3K1iq\" >For calculating the value of exposures for the purposes of Article 395(1) an institution may use the &apos;fully adjusted exposure value&apos;</p></content></subparagraph><subparagraph xml:id=\"imp_art_d1e34277_V6MXDb_BmKiVA\" ><content xml:id=\"imp_art_d1e34277_V6MXDb_sawpuz\" ><p xml:id=\"imp_art_d1e34277_V6MXDb_ixDMH6\" >as calculated under Part Three, Title II, Chapter 4 taking into account the credit risk mitigation, volatility adjustments, and any maturity mismatch (E*).</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e34277_V6MXDb\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e34277_V6MXDb_MG65v8\" >Article 4</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e34277_V6MXDb_Opeshn\" >Calculating the effect of the use of credit risk mitigation techniques</heading><paragraph xml:id=\"imp_art_d1e34277_V6MXDb_mdQK5X\" leos:origin=\"ec\"><num xml:id=\"imp_art_d1e34277_V6MXDb_rSp9HS\" leos:origin=\"ec\">1.</num><subparagraph leos:origin=\"ec\" leos:softaction=\"trans\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"transformed_imp_art_d1e34277_V6MXDb_mdQK5X\" ><content xml:id=\"imp_art_d1e34277_V6MXDb_g2YFMP\" ><p xml:id=\"imp_art_d1e34277_V6MXDb_A3K1iq\" >For calculating the value of exposures for the purposes of Article 395(1) an institution may use the &apos;fully adjusted exposure value&apos;</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e34277_V6MXDb_BmKiVA\" ><content xml:id=\"imp_art_d1e34277_V6MXDb_sawpuz\" ><p xml:id=\"imp_art_d1e34277_V6MXDb_ixDMH6\" >as calculated under Part Three, Title II, Chapter 4 taking into account the credit risk mitigation, volatility adjustments, and any maturity mismatch (E*).</p></content></subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_cke_article_ec_paragraph_cn_splitted() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\"   ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content></paragraph><paragraph xml:id=\"akn_art_para_1VF7qi\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T11:49:17.207+01:00\"><num xml:id=\"art_1_bM9Yvx\">1c.</num><subparagraph xml:id=\"art_1_vjLqth\" ><content xml:id=\"art_1_cC9aPz\" ><p xml:id=\"art_1_Qmw1fL\" >Paragraph</p></content></subparagraph><subparagraph xml:id=\"art_1_l2xM4D\" ><content xml:id=\"art_1_S92nAe\" ><p xml:id=\"art_1_PhwHwx\" >Text...</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\"   ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content></paragraph><paragraph xml:id=\"akn_art_para_1VF7qi\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T11:49:17.207+01:00\"><num xml:id=\"art_1_bM9Yvx\">1c.</num><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"art_1_vjLqth\" ><content xml:id=\"art_1_cC9aPz\" ><p xml:id=\"art_1_Qmw1fL\" >Paragraph</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"art_1_l2xM4D\" ><content xml:id=\"art_1_S92nAe\" ><p xml:id=\"art_1_PhwHwx\" >Text...</p></content></subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);

        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_cke_article_ec_subparagraph_ec_splitted() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e18032_19PFkP\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_UyZIpT\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_FPhfU3\" >Requirements to qualify for the treatment set out in Article 153(3)</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_MTUE8P\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_I6puhC\" >1.</num><subparagraph xml:id=\"imp_art_d1e18032_19PFkP_Db7137\" leos:origin=\"ec\"><content xml:id=\"imp_art_d1e18032_19PFkP_p1CpOV\" ><p xml:id=\"imp_art_d1e18032_19PFkP_ivbEGy\" >To be eligible for the treatment set out in Article 153(3), credit protection deriving from a guarantee or credit derivative shall meet</p></content></subparagraph><subparagraph xml:id=\"imp_art_d1e18032_19PFkP_UeBnhG\" ><content xml:id=\"imp_art_d1e18032_19PFkP_dmjmAH\" ><p xml:id=\"imp_art_d1e18032_19PFkP_tpSrWd\" >the following conditions:</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e18032_19PFkP\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_UyZIpT\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_FPhfU3\" >Requirements to qualify for the treatment set out in Article 153(3)</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_MTUE8P\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e18032_19PFkP_I6puhC\" >1.</num><subparagraph xml:id=\"imp_art_d1e18032_19PFkP_Db7137\" leos:origin=\"ec\"><content xml:id=\"imp_art_d1e18032_19PFkP_p1CpOV\" ><p xml:id=\"imp_art_d1e18032_19PFkP_ivbEGy\" >To be eligible for the treatment set out in Article 153(3), credit protection deriving from a guarantee or credit derivative shall meet</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e18032_19PFkP_UeBnhG\" ><content xml:id=\"imp_art_d1e18032_19PFkP_dmjmAH\" ><p xml:id=\"imp_art_d1e18032_19PFkP_tpSrWd\" >the following conditions:</p></content></subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_cke_article_ec_subparagraph_cn_splitted() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" leos:softuser=\"demo\" ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph xml:id=\"art_1__para_1\" leos:origin=\"ec\" leos:softuser=\"demo\"><num xml:id=\"art_1__para_1__num\" leos:origin=\"ec\">1.</num><subparagraph leos:origin=\"ec\" leos:softaction=\"trans\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.035+01:00\" xml:id=\"transformed_art_1__para_1\" ><content xml:id=\"art_1_lP7CeV\" ><p xml:id=\"art_1_wyGWnb\" >Text... <inline xml:id=\"art_1_0miExA\" leos:origin=\"ec\" name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline> sa</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.037+01:00\" xml:id=\"art_1_e0303i\" ><content xml:id=\"art_1_0SOAi3\" ><p xml:id=\"art_1_NPExjj\" >as as as</p></content></subparagraph></paragraph>              <paragraph xml:id=\"akn_art_para_OQTUpc\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:22:42.155+01:00\"><num xml:id=\"art_1_hNZU6V\">1a.</num><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.055+01:00\" xml:id=\"art_1_Bw0s3c\" ><content xml:id=\"art_1_WwmL3d\" ><p xml:id=\"art_1_0ciZrg\" >Paragraph</p></content></subparagraph><subparagraph xml:id=\"art_1_mUjRT9\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.070+01:00\"><content xml:id=\"art_1_WmFdLo\" ><p xml:id=\"art_1_kXVvwh\" >For calculating the value of exposures for the purposes of Article 395(1) an institution may use the &apos;fully adjusted exposure value&apos; as calculated under Part Three, Title II, Chapter 4 taking into account the credit risk mitigation,</p></content></subparagraph><subparagraph xml:id=\"art_1_PXd1Bx\" ><content xml:id=\"art_1_JlttGn\" ><p xml:id=\"art_1_pIRiM5\" >volatility adjustments, and any maturity mismatch (E*).</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" leos:softuser=\"demo\" ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph xml:id=\"art_1__para_1\" leos:origin=\"ec\" leos:softuser=\"demo\"><num xml:id=\"art_1__para_1__num\" leos:origin=\"ec\">1.</num><subparagraph leos:origin=\"ec\" leos:softaction=\"trans\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.035+01:00\" xml:id=\"transformed_art_1__para_1\" ><content xml:id=\"art_1_lP7CeV\" ><p xml:id=\"art_1_wyGWnb\" >Text... <inline xml:id=\"art_1_0miExA\" leos:origin=\"ec\" name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline> sa</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.037+01:00\" xml:id=\"art_1_e0303i\" ><content xml:id=\"art_1_0SOAi3\" ><p xml:id=\"art_1_NPExjj\" >as as as</p></content></subparagraph></paragraph>              <paragraph xml:id=\"akn_art_para_OQTUpc\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:22:42.155+01:00\"><num xml:id=\"art_1_hNZU6V\">1a.</num><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.055+01:00\" xml:id=\"art_1_Bw0s3c\" ><content xml:id=\"art_1_WwmL3d\" ><p xml:id=\"art_1_0ciZrg\" >Paragraph</p></content></subparagraph><subparagraph xml:id=\"art_1_mUjRT9\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:36:13.070+01:00\"><content xml:id=\"art_1_WmFdLo\" ><p xml:id=\"art_1_kXVvwh\" >For calculating the value of exposures for the purposes of Article 395(1) an institution may use the &apos;fully adjusted exposure value&apos; as calculated under Part Three, Title II, Chapter 4 taking into account the credit risk mitigation,</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"art_1_PXd1Bx\" ><content xml:id=\"art_1_JlttGn\" ><p xml:id=\"art_1_pIRiM5\" >volatility adjustments, and any maturity mismatch (E*).</p></content></subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_cke_article_cn_added_paragraph() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_x6DzYY\" leos:origin=\"cn\" leos:editable=\"true\" leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.686+01:00\"><num xml:id=\"akn_art_x6DzYY_1SAdsJ\">Article -1</num><heading xml:id=\"akn_art_x6DzYY_4athz2\">Article heading...</heading><paragraph xml:id=\"akn_art_x6DzYY-par1\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.702+01:00\"><num xml:id=\"akn_art_x6DzYY_64P9pb\">1.</num><content xml:id=\"akn_art_x6DzYY_4mTUzb\"><p xml:id=\"akn_art_x6DzYY_AthVYm\">Text...</p></content></paragraph><paragraph xml:id=\"akn_art_x6DzYY_OjZlKx\"  leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.702+01:00\"><num xml:id=\"akn_art_x6DzYY_2XWlXQ\" >2.</num><content xml:id=\"akn_art_x6DzYY_H4aFhJ\" ><p xml:id=\"akn_art_x6DzYY_PTKHJy\" >Test</p></content></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_x6DzYY\" leos:origin=\"cn\" leos:editable=\"true\" leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.686+01:00\"><num xml:id=\"akn_art_x6DzYY_1SAdsJ\">Article -1</num><heading xml:id=\"akn_art_x6DzYY_4athz2\">Article heading...</heading><paragraph xml:id=\"akn_art_x6DzYY-par1\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.702+01:00\"><num xml:id=\"akn_art_x6DzYY_64P9pb\">1.</num><content xml:id=\"akn_art_x6DzYY_4mTUzb\"><p xml:id=\"akn_art_x6DzYY_AthVYm\">Text...</p></content></paragraph><paragraph leos:origin=\"cn\" xml:id=\"akn_art_x6DzYY_OjZlKx\"  leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:16:11.702+01:00\"><num xml:id=\"akn_art_x6DzYY_2XWlXQ\" >2.</num><content xml:id=\"akn_art_x6DzYY_H4aFhJ\" ><p xml:id=\"akn_art_x6DzYY_PTKHJy\" >Test</p></content></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_cke_article_cn_paragraph_splitted() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_9WnbxA\" leos:origin=\"cn\" leos:editable=\"true\" leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:18:35.254+01:00\"><num xml:id=\"akn_art_9WnbxA_dKRiJn\">Article -2</num><heading xml:id=\"akn_art_9WnbxA_re50lT\">Article heading...</heading><paragraph xml:id=\"akn_art_9WnbxA-par1\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:22:35.303+01:00\"><num xml:id=\"akn_art_9WnbxA_X3ieIj\">1.</num><subparagraph xml:id=\"akn_art_9WnbxA_I6Rser\" ><content xml:id=\"akn_art_9WnbxA_ONy1Cf\" ><p xml:id=\"akn_art_9WnbxA_wijfWd\" >Text...</p></content></subparagraph><subparagraph xml:id=\"akn_art_9WnbxA_Z8lLi7\" ><content xml:id=\"akn_art_9WnbxA_pVuh3B\" ><p xml:id=\"akn_art_9WnbxA_vDfgXM\" >test</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_9WnbxA\" leos:origin=\"cn\" leos:editable=\"true\" leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:18:35.254+01:00\"><num xml:id=\"akn_art_9WnbxA_dKRiJn\">Article -2</num><heading xml:id=\"akn_art_9WnbxA_re50lT\">Article heading...</heading><paragraph xml:id=\"akn_art_9WnbxA-par1\" leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:22:35.303+01:00\"><num xml:id=\"akn_art_9WnbxA_X3ieIj\">1.</num><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"akn_art_9WnbxA_I6Rser\" ><content xml:id=\"akn_art_9WnbxA_ONy1Cf\" ><p xml:id=\"akn_art_9WnbxA_wijfWd\" >Text...</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"akn_art_9WnbxA_Z8lLi7\" ><content xml:id=\"akn_art_9WnbxA_pVuh3B\" ><p xml:id=\"akn_art_9WnbxA_vDfgXM\" >test</p></content></subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_toc_in_article_ec_empty_paragraph_ec_added_subparagraph() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" leos:softuser=\"demo\"><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\" leos:softuser=\"demo\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num>              <subparagraph xml:id=\"akn_art_subpara_HaUTRs\" leos:softaction=\"add\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:56:51.605+01:00\"><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content>              </subparagraph>              <subparagraph xml:id=\"akn_art_subpara_JwqJAz\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:56:51.603+01:00\" leos:origin=\"cn\">                <content xml:id=\"art_1_dB5lte\" >                  <p xml:id=\"art_1_WxnFDr\" >Subparagraph Text...</p>                </content>              </subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" leos:softuser=\"demo\"><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\" leos:softuser=\"demo\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num>              <subparagraph leos:origin=\"ec\" xml:id=\"transformed_art_1__para_1\" leos:softaction=\"trans\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:56:51.605+01:00\"><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content>              </subparagraph>              <subparagraph xml:id=\"akn_art_subpara_JwqJAz\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T11:56:51.603+01:00\" leos:origin=\"cn\">                <content xml:id=\"art_1_dB5lte\" >                  <p xml:id=\"art_1_WxnFDr\" >Subparagraph Text...</p>                </content>              </subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_toc_in_article_ec_empty_paragraph_cn_added_subparagraph() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content></paragraph><paragraph xml:id=\"akn_art_para_nf5uS5\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:29.283+01:00\" leos:origin=\"cn\"><num xml:id=\"art_1_tXlaEH\"  leos:editable=\"false\">1a.</num>              <subparagraph xml:id=\"akn_art_subpara_7oRqVE\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:52.692+01:00\"><content xml:id=\"art_1_IcSoZv\" >                  <p xml:id=\"art_1_vJRdj0\" >Paragraph Text...</p>                </content>              </subparagraph>              <subparagraph xml:id=\"akn_art_subpara_egDbOK\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:52.692+01:00\" leos:origin=\"cn\">                <content xml:id=\"art_1_ZJYR3v\" >                  <p xml:id=\"art_1_2Evnrd\" >Subparagraph Text...</p>                </content>              </subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" xml:id=\"art_1\" leos:editable=\"false\" leos:deletable=\"false\" ><num leos:origin=\"ec\" leos:editable=\"false\" xml:id=\"art_1__num\">Article 1</num><heading leos:origin=\"ec\" xml:id=\"art_1__heading\">Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"art_1__para_1\"><num leos:origin=\"ec\" xml:id=\"art_1__para_1__num\">1.</num><content leos:origin=\"ec\" xml:id=\"art_1__para_1__content\"><p leos:origin=\"ec\" xml:id=\"art_1__para_1__content__p\">Text... <inline leos:origin=\"ec\" xml:id=\"art_1_0miExA\"  name=\"math-tex\">\\(x = {-b \\pm \\sqrt{b^2-4ac} \\over 2a}\\)</inline></p></content></paragraph><paragraph xml:id=\"akn_art_para_nf5uS5\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:29.283+01:00\" leos:origin=\"cn\"><num xml:id=\"art_1_tXlaEH\"  leos:editable=\"false\">1a.</num>              <subparagraph leos:origin=\"cn\" xml:id=\"akn_art_subpara_7oRqVE\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:52.692+01:00\"><content xml:id=\"art_1_IcSoZv\" >                  <p xml:id=\"art_1_vJRdj0\" >Paragraph Text...</p>                </content>              </subparagraph>              <subparagraph xml:id=\"akn_art_subpara_egDbOK\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-24T10:44:52.692+01:00\" leos:origin=\"cn\">                <content xml:id=\"art_1_ZJYR3v\" >                  <p xml:id=\"art_1_2Evnrd\" >Subparagraph Text...</p>                </content>              </subparagraph></paragraph></article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_updateNewElements_toc_added_article_cn() throws Exception {
        String xmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_hbaALO\" leos:editable=\"true\"  leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:08:32.247+01:00\" leos:origin=\"cn\"><num xml:id=\"akn_art_hbaALO_1Onb1M\"  leos:editable=\"false\">Article -1</num>                            <heading xml:id=\"akn_art_hbaALO_OMUNZD\" >Article heading...</heading>              <paragraph xml:id=\"akn_art_hbaALO-par1\">                <num xml:id=\"akn_art_hbaALO_eu7RzT\" >1.</num>                <content xml:id=\"akn_art_hbaALO_p78bjk\" >                  <p xml:id=\"akn_art_hbaALO_qKD1jn\" >Text...</p>                </content>              </paragraph>            </article>" +
                "</root>";

        String expectedXmlContent = "<root xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article xml:id=\"akn_art_hbaALO\" leos:editable=\"true\"  leos:deletable=\"true\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"2019-01-29T12:08:32.247+01:00\" leos:origin=\"cn\"><num xml:id=\"akn_art_hbaALO_1Onb1M\"  leos:editable=\"false\">Article -1</num>                            <heading xml:id=\"akn_art_hbaALO_OMUNZD\" >Article heading...</heading>              <paragraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"akn_art_hbaALO-par1\">                <num xml:id=\"akn_art_hbaALO_eu7RzT\" >1.</num>                <content xml:id=\"akn_art_hbaALO_p78bjk\" >                  <p xml:id=\"akn_art_hbaALO_qKD1jn\" >Text...</p>                </content>              </paragraph>            </article>" +
                "</root>";

        VTDNav vtdNav = VTDUtils.setupVTDNav(xmlContent.getBytes(UTF_8));
        XMLModifier xmlModifier = new XMLModifier(vtdNav);

        vtdXmlContentProcessor.specificInstanceXMLPostProcessing(xmlModifier);

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        xmlModifier.output(baos);
        assertEquals(expectedXmlContent, baos.toString("UTF-8"));
    }

    @Test
    public void test_merge_paragraph_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_w7XSAU\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_jzKMrH\" >2.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_3eweeF\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_0od3XF\" >Where reference in this Regulation is made to real estate or residential or commercial immovable property or a mortgage on such property, it shall include shares in Finnish residential housing companies operating in accordance with the Finnish Housing Company Act of 1991 or subsequent equivalent legislation. Member States or their competent authorities may allow shares constituting an equivalent indirect holding of real estate to be treated as a direct holding of real estate provided that such an indirect holding is specifically regulated in the national law of the Member State concerned and that, when pledged as collateral, it provides equivalent protection to creditors.</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OzDWGj\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_CcB0QB\" >3.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_zO3WCu\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_9ewOjj\" >Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_w7XSAU\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_jzKMrH\" >2.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_3eweeF\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_0od3XF\" >Where reference in this Regulation is made to real estate or residential or commercial immovable property or a mortgage on such property, it shall include shares in Finnish residential housing companies operating in accordance with the Finnish Housing Company Act of 1991 or subsequent equivalent legislation. Member States or their competent authorities may allow shares constituting an equivalent indirect holding of real estate to be treated as a direct holding of real estate provided that such an indirect holding is specifically regulated in the national law of the Member State concerned and that, when pledged as collateral, it provides equivalent protection to creditors. Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_OzDWGj\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"del\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_CcB0QB\" >3.</num><content leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_zO3WCu\" ><p leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_9ewOjj\" >Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></paragraph></article>" +
                "</root>";

        String elementToMerge = "<paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OzDWGj\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_CcB0QB\" >3.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_zO3WCu\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_9ewOjj\" >Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></paragraph>";
        byte[] result = vtdXmlContentProcessor.mergeElement(xmlContent.getBytes(), elementToMerge, PARAGRAPH, "imp_art_d1e1276_XAPlpX_OzDWGj");

        assertEquals(expectedXmlContent, new String(result, UTF_8));
    }

    @Test
    public void test_merge_back_splitted_paragraph_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_w7XSAU\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_jzKMrH\" >2.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_3eweeF\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_0od3XF\" >Where reference in this Regulation is made to real estate or residential or commercial immovable property or a mortgage on such property, it shall include shares in Finnish residential housing companies operating in accordance with the Finnish Housing Company Act of 1991 or subsequent equivalent legislation. Member States or their competent authorities may allow shares constituting an equivalent indirect holding of real estate to be treated as a direct holding of real estate provided that such an indirect holding is specifically regulated in the national law of the Member State concerned and that, when pledged as collateral, it provides equivalent protection to creditors.</p></content></paragraph><paragraph xml:id=\"imp_art_d1e1276_XAPlpX_OzDWGj\" leos:origin=\"ec\"><num xml:id=\"imp_art_d1e1276_XAPlpX_CcB0QB\" leos:origin=\"ec\">3.</num><subparagraph leos:origin=\"ec\" leos:softaction=\"trans\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"transformed_imp_art_d1e1276_XAPlpX_OzDWGj\" ><content xml:id=\"imp_art_d1e1276_XAPlpX_zO3WCu\"><p xml:id=\"imp_art_d1e1276_XAPlpX_9ewOjj\">Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the</p></content></subparagraph><subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e1276_XAPlpX_to5uqp\" ><content xml:id=\"imp_art_d1e1276_XAPlpX_jhLJhj\" ><p xml:id=\"imp_art_d1e1276_XAPlpX_5wgkKS\" >event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></subparagraph></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_w7XSAU\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_jzKMrH\" >2.</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_3eweeF\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_0od3XF\" >Where reference in this Regulation is made to real estate or residential or commercial immovable property or a mortgage on such property, it shall include shares in Finnish residential housing companies operating in accordance with the Finnish Housing Company Act of 1991 or subsequent equivalent legislation. Member States or their competent authorities may allow shares constituting an equivalent indirect holding of real estate to be treated as a direct holding of real estate provided that such an indirect holding is specifically regulated in the national law of the Member State concerned and that, when pledged as collateral, it provides equivalent protection to creditors.</p></content></paragraph><paragraph xml:id=\"imp_art_d1e1276_XAPlpX_OzDWGj\" leos:origin=\"ec\" leos:softtrans_from=\"transformed_imp_art_d1e1276_XAPlpX_OzDWGj\"><num xml:id=\"imp_art_d1e1276_XAPlpX_CcB0QB\" leos:origin=\"ec\">3.</num><content xml:id=\"imp_art_d1e1276_XAPlpX_zO3WCu\"><p xml:id=\"imp_art_d1e1276_XAPlpX_9ewOjj\">Trade finance as referred to in point (80) of paragraph 1 is generally uncommitted and requires satisfactory supporting transactional documentation for each drawdown request enabling refusal of the finance in the event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></paragraph></article>" +
                "</root>";

        String elementToMerge = "<subparagraph leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e1276_XAPlpX_to5uqp\" ><content xml:id=\"imp_art_d1e1276_XAPlpX_jhLJhj\" ><p xml:id=\"imp_art_d1e1276_XAPlpX_5wgkKS\" >event of any doubt about credit-worthiness or the supporting transactional documentation. Repayment of trade finance exposures is usually independent of the borrower, the funds instead coming from cash received from importers or resulting from proceeds of the sales of the underlying goods.</p></content></subparagraph>";
        byte[] result = vtdXmlContentProcessor.mergeElement(xmlContent.getBytes(), elementToMerge, SUBPARAGRAPH, "imp_art_d1e1276_XAPlpX_to5uqp");

        assertEquals(expectedXmlContent, new String(result, UTF_8));
    }

    @Test
    public void test_merge_points_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1221_qiqjdt\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xISt2k\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_P9hekz\" >Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_B90HX5\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GxCek2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xyt4AQ\" >This Regulation lays down uniform rules concerning general prudential requirements that institutions supervised under Directive 2013/36/EU shall comply with in relation to the following items:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_WApFyt\" ><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_Z4LsCi\" ><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NgnbOF\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_2jwLTy\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_LEPz3S\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_s6AuCN\" >own funds requirements relating to entirely quantifiable, uniform and standardised elements of credit risk, market risk, operational risk and settlement risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_YmrR20\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NjFXxB\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_zCvTSP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_I1evcQ\" >requirements limiting large exposures;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_OSeAWL\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_JACup8\" >(c)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rAY5q2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_vTMQL9\" >after the delegated act referred to in Article 460 has entered into force, liquidity requirements relating to entirely quantifiable, uniform and standardised elements of liquidity risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_oaz8D1\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NkI3nS\" >(d)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_3bM3YV\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_r3HIw7\" >reporting requirements related to points (a), (b) and (c) and to leverage;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GEj1jJ\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rlOyG4\" >(e)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_iF4nRP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_64XI6K\" >public disclosure requirements.</p></content></point></list></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1221_qiqjdt\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xISt2k\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_P9hekz\" >Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_B90HX5\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GxCek2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xyt4AQ\" >This Regulation lays down uniform rules concerning general prudential requirements that institutions supervised under Directive 2013/36/EU shall comply with in relation to the following items:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_WApFyt\" ><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_Z4LsCi\" ><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NgnbOF\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_2jwLTy\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_LEPz3S\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_s6AuCN\" >own funds requirements relating to entirely quantifiable, uniform and standardised elements of credit risk, market risk, operational risk and settlement risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_YmrR20\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NjFXxB\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_zCvTSP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_I1evcQ\" >requirements limiting large exposures;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_OSeAWL\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_JACup8\" >(c)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rAY5q2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_vTMQL9\" >after the delegated act referred to in Article 460 has entered into force, liquidity requirements relating to entirely quantifiable, uniform and standardised elements of liquidity risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_oaz8D1\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NkI3nS\" >(d)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_3bM3YV\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_r3HIw7\" >reporting requirements related to points (a), (b) and (c) and to leverage; public disclosure requirements.</p></content></point><point leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1221_qiqjdt_GEj1jJ\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"del\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1221_qiqjdt_rlOyG4\" >(e)</num><content leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1221_qiqjdt_iF4nRP\" ><p leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1221_qiqjdt_64XI6K\" >public disclosure requirements.</p></content></point></list></paragraph></article>" +
                "</root>";

        String elementToMerge = "<point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GEj1jJ\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rlOyG4\" >(e)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_iF4nRP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_64XI6K\" >public disclosure requirements.</p></content></point>";
        byte[] result = vtdXmlContentProcessor.mergeElement(xmlContent.getBytes(), elementToMerge, POINT, "imp_art_d1e1221_qiqjdt_GEj1jJ");

        assertEquals(expectedXmlContent, new String(result, UTF_8));
    }

    @Test
    public void test_merge_back_splitted_point_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1221_qiqjdt\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xISt2k\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_P9hekz\" >Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_B90HX5\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GxCek2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xyt4AQ\" >This Regulation lays down uniform rules concerning general prudential requirements that institutions supervised under Directive 2013/36/EU shall comply with in relation to the following items:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_WApFyt\" ><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_Z4LsCi\" ><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NgnbOF\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_2jwLTy\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_LEPz3S\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_s6AuCN\" >own funds requirements relating to entirely quantifiable, uniform and standardised elements of credit risk, market risk, operational risk and settlement risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_YmrR20\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NjFXxB\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_zCvTSP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_I1evcQ\" >requirements limiting large exposures;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_OSeAWL\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_JACup8\" >(c)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rAY5q2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_vTMQL9\" >after the delegated act referred to in Article 460 has entered into force, liquidity requirements relating to entirely quantifiable, uniform and standardised elements of liquidity risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_oaz8D1\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NkI3nS\" >(d)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_3bM3YV\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_r3HIw7\" >reporting requirements related to points (a), (b) and (c) and to leverage;</p></content></point><point xml:id=\"imp_art_d1e1221_qiqjdt_GEj1jJ\" leos:origin=\"ec\"><num xml:id=\"imp_art_d1e1221_qiqjdt_rlOyG4\" leos:origin=\"ec\">(e)</num><alinea leos:origin=\"ec\" leos:softaction=\"trans\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"transformed_imp_art_d1e1221_qiqjdt_GEj1jJ\" ><content xml:id=\"imp_art_d1e1221_qiqjdt_iF4nRP\"><p xml:id=\"imp_art_d1e1221_qiqjdt_64XI6K\">public</p></content></alinea><alinea leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e1221_qiqjdt_ajnCHE\" ><content xml:id=\"imp_art_d1e1221_qiqjdt_mgfYGC\" ><p xml:id=\"imp_art_d1e1221_qiqjdt_XvE23L\" >disclosure requirements.</p></content></alinea></point></list></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1221_qiqjdt\"><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xISt2k\" >Article 2</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_P9hekz\" >Scope</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_B90HX5\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_GxCek2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_xyt4AQ\" >This Regulation lays down uniform rules concerning general prudential requirements that institutions supervised under Directive 2013/36/EU shall comply with in relation to the following items:</p></content></paragraph><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_WApFyt\" ><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_Z4LsCi\" ><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NgnbOF\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_2jwLTy\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_LEPz3S\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_s6AuCN\" >own funds requirements relating to entirely quantifiable, uniform and standardised elements of credit risk, market risk, operational risk and settlement risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_YmrR20\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NjFXxB\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_zCvTSP\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_I1evcQ\" >requirements limiting large exposures;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_OSeAWL\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_JACup8\" >(c)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_rAY5q2\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_vTMQL9\" >after the delegated act referred to in Article 460 has entered into force, liquidity requirements relating to entirely quantifiable, uniform and standardised elements of liquidity risk;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_oaz8D1\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_NkI3nS\" >(d)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_3bM3YV\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1221_qiqjdt_r3HIw7\" >reporting requirements related to points (a), (b) and (c) and to leverage;</p></content></point><point xml:id=\"imp_art_d1e1221_qiqjdt_GEj1jJ\" leos:origin=\"ec\" leos:softtrans_from=\"transformed_imp_art_d1e1221_qiqjdt_GEj1jJ\"><num xml:id=\"imp_art_d1e1221_qiqjdt_rlOyG4\" leos:origin=\"ec\">(e)</num><content xml:id=\"imp_art_d1e1221_qiqjdt_iF4nRP\"><p xml:id=\"imp_art_d1e1221_qiqjdt_64XI6K\">public disclosure requirements.</p></content></point></list></paragraph></article>" +
                "</root>";

        String elementToMerge = "<alinea leos:origin=\"cn\" leos:softaction=\"add\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\" xml:id=\"imp_art_d1e1221_qiqjdt_ajnCHE\" ><content xml:id=\"imp_art_d1e1221_qiqjdt_mgfYGC\" ><p xml:id=\"imp_art_d1e1221_qiqjdt_XvE23L\" >disclosure requirements.</p></content></alinea>";
        byte[] result = vtdXmlContentProcessor.mergeElement(xmlContent.getBytes(), elementToMerge, SUBPOINT, "imp_art_d1e1221_qiqjdt_ajnCHE");

        assertEquals(expectedXmlContent, new String(result, UTF_8));
    }

    @Test
    public void test_merge_softmoved_point_into_no_softmoved_point_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><subparagraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_KyB4qy\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></subparagraph><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_J7TbDh\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMwJok\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_TD9PRk\" >(2)</num><alinea leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vPsN8G\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Atnkut\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Z5vpZD\" >'investment firm' means a person as defined in point (1) of Article 4(1) of Directive 2004/39/EC, which is subject to the requirements imposed by that Directive, excluding the following:</p></content></alinea><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aZi7xk\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_X9uQs2\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_G73ZDv\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMkDsZ\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_oKMA14\" >credit institutions;</p></content></point><point leos:softmove_label=\"MOVED from point (c)\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:softaction=\"move_from\" leos:softactionroot=\"true\" leos:softmove_from=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"cn\" xml:id=\"imp_art_d1e1276_XAPlpX_jh6oZJ\" >(aa)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bD74NR\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Sk62PF\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4sLVEG\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vpwYsL\" >local firms;</p></content></point><point leos:softmove_label=\"MOVED to point (aa)\" leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"move_to\" leos:softactionroot=\"true\" leos:softmove_to=\"imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_jh6oZJ\" >(c)</num><content leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point></list></point></list></paragraph></article>" +
                "</root>";

        String expectedXmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:affected=\"true\" leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:affected=\"true\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><subparagraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_KyB4qy\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></subparagraph><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_J7TbDh\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point><point leos:affected=\"true\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMwJok\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_TD9PRk\" >(2)</num><alinea leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vPsN8G\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Atnkut\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Z5vpZD\" >'investment firm' means a person as defined in point (1) of Article 4(1) of Directive 2004/39/EC, which is subject to the requirements imposed by that Directive, excluding the following:</p></content></alinea><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aZi7xk\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_X9uQs2\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_G73ZDv\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMkDsZ\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_oKMA14\" >credit institutions; firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bD74NR\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Sk62PF\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4sLVEG\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vpwYsL\" >local firms;</p></content></point><point leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"del\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_jh6oZJ\" >(c)</num><content leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point></list></point></list></paragraph></article>" +
                "</root>";

        String elementToMerge = "<point leos:softmove_label=\"MOVED from point (c)\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:softaction=\"move_from\" leos:softactionroot=\"true\" leos:softmove_from=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"cn\" xml:id=\"imp_art_d1e1276_XAPlpX_jh6oZJ\" >(aa)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point>";
        byte[] result = vtdXmlContentProcessor.mergeElement(xmlContent.getBytes(), elementToMerge, POINT, "imp_art_d1e1276_XAPlpX_4K2GdA");

        assertEquals(expectedXmlContent, new String(result, UTF_8));
    }

    @Test
    public void test_merge_first_point_of_list_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><subparagraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_KyB4qy\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></subparagraph><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_J7TbDh\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMwJok\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_TD9PRk\" >(2)</num><alinea leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vPsN8G\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Atnkut\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Z5vpZD\" >'investment firm' means a person as defined in point (1) of Article 4(1) of Directive 2004/39/EC, which is subject to the requirements imposed by that Directive, excluding the following:</p></content></alinea><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aZi7xk\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_X9uQs2\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_G73ZDv\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMkDsZ\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_oKMA14\" >credit institutions;</p></content></point><point leos:softmove_label=\"MOVED from point (c)\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:softaction=\"move_from\" leos:softactionroot=\"true\" leos:softmove_from=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"cn\" xml:id=\"imp_art_d1e1276_XAPlpX_jh6oZJ\" >(aa)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bD74NR\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Sk62PF\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4sLVEG\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vpwYsL\" >local firms;</p></content></point><point leos:softmove_label=\"MOVED to point (aa)\" leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"move_to\" leos:softactionroot=\"true\" leos:softmove_to=\"imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_jh6oZJ\" >(c)</num><content leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point></list></point></list></paragraph></article>" +
                "</root>";

        String elementToMerge = "<point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point>";
        String[] result = vtdXmlContentProcessor.getMergeOnElement(xmlContent.getBytes(), elementToMerge, POINT, "imp_art_d1e1276_XAPlpX_VKPtnz");

        assertNull(result);
    }

    @Test
    public void test_merge_point_with_softdeleted_ec() throws Exception {
        String xmlContent = "<root xml:id=\"akn_f2ejkO\" xmlns:leos=\"urn:eu:europa:ec:leos\" xmlns:xml=\"http://www.w3.org/XML/1998/namespace\">" +
                "<article leos:origin=\"ec\" leos:editable=\"false\"  leos:deletable=\"false\" xml:id=\"imp_art_d1e1276_XAPlpX\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_PCoY8I\" >Article 5</num><heading leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_yDHX83\" >Definitions</heading><paragraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_494qUn\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_gxZ46e\" >1.</num><subparagraph leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_KyB4qy\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_nBJK4i\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aW9qa6\" >For the purposes of this Regulation, the following definitions shall apply:</p></content></subparagraph><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_J7TbDh\"><point leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"del\" leos:softactionroot=\"true\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_jh6oZJ\" >(c)</num><content leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"deleted_imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMwJok\"  ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_TD9PRk\" >(2)</num><alinea leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vPsN8G\" ><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Atnkut\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Z5vpZD\" >'investment firm' means a person as defined in point (1) of Article 4(1) of Directive 2004/39/EC, which is subject to the requirements imposed by that Directive, excluding the following:</p></content></alinea><list leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_aZi7xk\"><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_X9uQs2\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_G73ZDv\" >(a)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_HMkDsZ\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_oKMA14\" >credit institutions;</p></content></point><point leos:softmove_label=\"MOVED from point (c)\" leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:softaction=\"move_from\" leos:softactionroot=\"true\" leos:softmove_from=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"cn\" xml:id=\"imp_art_d1e1276_XAPlpX_jh6oZJ\" >(aa)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point><point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bD74NR\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_Sk62PF\" >(b)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_4sLVEG\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_vpwYsL\" >local firms;</p></content></point><point leos:softmove_label=\"MOVED to point (aa)\" leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_4K2GdA\"  leos:editable=\"false\" leos:deletable=\"false\" leos:softaction=\"move_to\" leos:softactionroot=\"true\" leos:softmove_to=\"imp_art_d1e1276_XAPlpX_4K2GdA\" leos:softuser=\"demo\" leos:softdate=\"1900-01-01T00:00:00.000+01:00\"><num leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_jh6oZJ\" >(c)</num><content leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_fYqac8\" ><p leos:origin=\"ec\" xml:id=\"moved_imp_art_d1e1276_XAPlpX_Eif3cN\" >firms which are not authorised to provide the ancillary service referred to in point (1) of Section B of Annex I to Directive 2004/39/EC, which provide only one or more of the investment services and activities listed in points 1, 2, 4 and 5 of Section A of Annex I to that Directive, and which are not permitted to hold money or securities belonging to their clients and which for that reason may not at any time place themselves in debt with those clients;</p></content></point></list></point></list></paragraph></article>" +
                "</root>";

        String elementToMerge = "<point leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_VKPtnz\" ><num leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_bx6xKW\" >(1)</num><content leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_syrXR4\" ><p leos:origin=\"ec\" xml:id=\"imp_art_d1e1276_XAPlpX_OpsDCb\" >'credit institution' means an undertaking the business of which is to take deposits or other repayable funds from the public and to grant credits for its own account;</p></content></point>";
        String[] result = vtdXmlContentProcessor.getMergeOnElement(xmlContent.getBytes(), elementToMerge, POINT, "imp_art_d1e1276_XAPlpX_VKPtnz");

        assertNull(result);
    }

}
